import numpy as np
import pandas as pd 
import torch
import pickle

from mushroom_rl.algorithms.value import DQN, DoubleDQN, AveragedDQN
from mushroom_rl.algorithms.actor_critic import A2C, DDPG, TD3, SAC, TRPO, PPO
from mushroom_rl.core import Core
from mushroom_rl.environments import *
from mushroom_rl.policy import EpsGreedy, Boltzmann, GaussianTorchPolicy, OrnsteinUhlenbeckPolicy
from mushroom_rl.approximators.parametric.torch_approximator import TorchApproximator
from mushroom_rl.utils.dataset import compute_J
from mushroom_rl.utils.parameters import Parameter, LinearParameter, ExponentialParameter

from RL_for_reco.Item_Reco import Item_Reco
from RL_for_reco.Network_for_Reco import Network_for_Reco, TorchApproximator_cuda

ALG_NAMES = {'DQN': DQN, 'DDQN': DoubleDQN, 'ADQN': AveragedDQN,   ## value-based
             'A2C': A2C, 'DDPG': DDPG, 'TD3': TD3, 'SAC': SAC, 'TRPO': TRPO, 'PPO': PPO}   ## policy-based
PI_PR_NAMES = {'Static': Parameter, 'Linear': LinearParameter, 'Exp': ExponentialParameter}  ## only for value-based
PI_NAMES = {'EG': EpsGreedy, 'BTM': Boltzmann, 
            'GSTorch': GaussianTorchPolicy, 'OUNoise': OrnsteinUhlenbeckPolicy}
ENV_NAMES = {'IR': Item_Reco}

'''
** Policy Gradient **
- A2C, PPO: need policy model, 1 critic network and 1 actor optimizer
- TRPO: need policy model and 1 critic network with max_kl
- DDPG, TD3: need (noise) policy model, 1 critic network, 1 actor network and 1 actor optimizer
- SAC: 2 actor networks and 1 critic network and 1 actor optimizer
'''


class RL_Learn:
    def __init__(self, env_name, alg_name, pi_pr_name=None, pi_name=None, env_params={}, pi_pr_params={}, alg_params={}):
        """
        Constructor

        env_name: only IR
        alg_name: one of {'DQN': DQN, 'DDQN': DoubleDQN, 'ADQN': AveragedDQN,   ## value-based
                          'A2C': A2C, 'DDPG': DDPG, 'TD3': TD3, 'SAC': SAC, 'TRPO': TRPO, 'PPO': PPO}   ## policy-based
        pi_pr_name(default=None): one of {'Static': Parameter, 'Linear': LinearParameter, 'Exp': ExponentialParameter}
        pi_name(default=None): one of {'EG': EpsGreedy, 'BTM': Boltzmann, ## value-based
                                       'GSTorch': GaussianTorchPolicy, ## policy-based
                                       'OUNoise': OrnsteinUhlenbeckPolicy}  ## add noise
        
        env_params: 
            items, array of item labels or integers representing items for value-based learning; 
                   array of a pair of min and max value over all items(actually all the pair are the same, in this version)
            gamma, learning rate
            horizon, length for agent to long in the environment
            trans_model_abs_path, absolute directory path where a transition model from (state, action) to (state, reward) is
            item_dist(default=None), a distribution item happening; when default, all the probability is the same as 1/|items|

        pi_pr_params:
            'Static'
                value, initial value
                min_value(default=None)
                max_value(default=None)
            'Linear'
                value, initial value
                threshold_value, some value to know (threshold_value - value)/n; 
                                 if which is smaller than 0, it is to be min value else to be max value
                n, denominator
            'Exp'
                value, initial value
                exp(default=1.0), used to be the power as value/(n ** exp) where n is the current number of being trained
                min_value(default=None)
                max_value(default=None)

            pi_name='EG'
                epsilon, one of 'Static', 'Linear' and 'Exp', the exploration coefficient
            pi_name='BTM'
                beta, one of 'Static', 'Linear' and 'Exp', the inverse of the temperature distribution; 
                      As the temperature approaches 0.0, the policy becomes more and more greedy
            pi_name='GSTorch'
                std_0(default=1.0), initial standard deviation
            pi_name='OUNoise'
                sigma, array of critic network output size, average magnitude of the random fluctuation per square-root time
                theta, rate of mean reversion
                dt, time interval
                x0(default=None), array of critic network output size, initial values of noise

        alg_params:
            ## params for Network_for_Reco
            hidden_dims, list/array of hidden unit sizes
            optimizer, dict, parameters to specify the optimizer algorithm
            loss, the loss function

            ### params for agent algorithms for DQN
            alg_name in ['DQN', 'DDQN', 'ADQN']
                batch_size, the number of samples in a batch
                target_update_frequency, the number of samples collected between each update of the target network
                replay_memory(default=None), one of ReplayMemory and PrioritizedReplayMemory, the object of the replay memory to use; 
                                             if None, a default replay memory is created
                initial_replay_size(default=500), the number of samples to collect before starting the learning
                max_replay_size(default=5000), the maximum number of samples in the replay memory
                n_approximators(default=1) the number of approximator for alg_name='ADQN'
                clip_reward(default=True), whether to clip the reward or not

            ### params for agent of actor-critic
            alg_name in ['A2C', 'PPO', 'DDPG', 'TD3', 'SAC']
                alg_name='A2C'
                    ent_coeff(default=0.0), coefficient for the entropy penalty
                    max_grad_norm(default=None), maximum norm for gradient clipping; 
                                                If None, no clipping will be performed, unless specified otherwise in actor_optimizer
                alg_name='PPO'
                    n_epochs_policy, number of policy updates for every dataset
                    batch_size, size of minibatches for every optimization step
                    eps_ppo, value for probability ratio clipping
                    lam(default=1.), lambda coefficient used by generalized advantage estimation
                alg_name in ['DDPG', 'TD3]
                    batch_size, the number of samples in a batch
                    initial_replay_size, the number of samples to collect before starting the learning
                    max_replay_size, the maximum number of samples in the replay memory
                    tau, value of coefficient for soft updates
                    policy_delay(default=1), the number of updates of the critic after which an actor update is implemented; 
                                             for 'TD3', default=2
                    alg_name='TD3'
                        noise_std(default=.2), standard deviation of the noise used for policy smoothing
                        noise_clip(default=.5), maximum absolute value for policy smoothing noise
                alg_name='SAC'
                    batch_size, the number of samples in a batch
                    initial_replay_size, the number of samples to collect before starting the learning
                    max_replay_size, the maximum number of samples in the replay memory
                    warmup_transitions, number of samples to accumulate in the replay memory to start the policy fitting
                    tau, value of coefficient for soft updates
                    lr_alpha, Learning rate for the entropy coefficient
                    target_entropy(default=None),target entropy for the policy; if None a default value is computed
            alg_name='TRPO'
                ent_coeff(default=0.0), coefficient for the entropy penalty
                max_kl(default=.001), maximum kl allowed for every policy update
                lam(default=1.), lambda coefficient used by generalized advantage estimation
                n_epochs_line_search(default=10), maximum number of iterations of the line search algorithm
                n_epochs_cg(default=10), maximum number of iterations of the conjugate gradient algorithm
                cg_damping(default=1e-2), damping factor for the conjugate gradient algorithm
                cg_residual_tol(default=1e-10), conjugate gradient residual tolerance
        """

        ## MDP
        self.env_name = ENV_NAMES[env_name]
        self.env = self.env_name(**env_params)
        

        ## Parameters of Agent(agent_params) and of Network_for_Reco(alg_params)
        self.agent_params = alg_params.copy()

        self.alg_params = {}
        for key in ['hidden_dims', 'optimizer', 'loss']:
            self.alg_params[key] = alg_params[key]
            if key == 'optimizer' and alg_name in ['A2C', 'PPO', 'DDPG', 'TD3', 'SAC']:
                self.agent_params.update({'actor_optimizer': alg_params[key]})
            del self.agent_params[key]
        self.alg_params['network'] = Network_for_Reco
        self.alg_params['input_shape'] = self.env.info.observation_space.shape
        if len(self.env.items.shape) == 1:
            self.alg_params['output_shape'] = self.env.info.action_space.size
            self.alg_params['n_actions'] = self.alg_params['output_shape'][0]
            self.alg_params['mode'] = 'q'
        else:
            self.alg_params['output_shape'] = (self.env.action_dim,)
            if alg_name in ['DDPG', 'TD3', 'SAC']:
                self.alg_params['mode'] = 'actor'

                self.crt_alg_params = self.alg_params.copy()
                del self.alg_params['optimizer']
                self.crt_alg_params['mode'] = 'critic'
                self.crt_alg_params['input_shape'] = (self.env.state_dim + self.env.action_dim,)
                self.crt_alg_params['output_shape'] = (1,)
            else:
                self.alg_params['mode'] = 'critic'
                self.alg_params['output_shape'] = (1,)

        # Policy 
        if pi_name is not None:
            self.pi_name = PI_NAMES[pi_name]
            if pi_name == 'OUNoise' and pi_pr_params is not None:
                self.pi_pr_params = pi_pr_params
            elif pi_name.endswith('Torch'):
                self.pi_pr_params = {
                    'use_cuda': False,
                    'network': Network_for_Reco,
                    'input_shape': self.env.info.observation_space.shape,
                    'output_shape': (self.env.action_dim,),
                    'hidden_dims': self.alg_params['hidden_dims']
                }
                self.pi_pr_params.update(pi_pr_params)
                self.policy = self.pi_name(**self.pi_pr_params)
            elif pi_pr_name is not None:
                self.pi_pr_name = PI_PR_NAMES[pi_pr_name]
                self.policy = self.pi_name(self.pi_pr_name(**pi_pr_params.copy()))


        self.agent_params['mdp_info'] = self.env.info
        if alg_name in ['DDPG', 'TD3']:
            self.agent_params['policy_class'] = self.pi_name
            self.agent_params['policy_params'] = self.pi_pr_params
            self.agent_params['actor_params'] = self.alg_params
            self.agent_params['critic_params'] = self.crt_alg_params
        elif alg_name == 'SAC':
            self.agent_params['critic_params'] = self.crt_alg_params
            self.agent_params['actor_mu_params'] = self.alg_params.copy()
            self.agent_params['actor_sigma_params'] = self.alg_params.copy()
        else:
            if alg_name != 'TRPO':
                self.agent_params['policy'] = self.policy
            
            if alg_name in ['DQN', 'DDQN', 'ADQN']:
                self.agent_params['approximator'] = TorchApproximator#_cuda
                self.agent_params['approximator_params'] = self.alg_params
            else:
                self.agent_params['critic_params'] = self.alg_params

        ## Agent and Core
        self.alg_name = ALG_NAMES[alg_name]
        self.agent = self.alg_name(**self.agent_params)
        self.core = Core(self.agent, self.env)

    def train(self, n_epochs, n_steps, train_frequency, print_epoch=True):
        for i in range(n_epochs):
            if print_epoch:
                print(f'---------- {i}th epoch -----------')
            self.core.learn(n_steps=n_steps, n_steps_per_fit=train_frequency)

    def compare_model_with_origin(self, initial_states, compared_rewards, n_samples=10000):
        if len(initial_states) > n_samples:
            idx = np.random.choice(range(len(initial_states)), n_samples, replace=False)
            samples = initial_states[idx]
            if compared_rewards is not None:
                raw_r = np.mean(np.array(compared_rewards)[idx])
        else:
            samples = initial_states
            if compared_rewards is not None:
                raw_r = np.mean(compared_rewards)
        dataset = self.core.evaluate(initial_states=samples)
        J = compute_J(dataset, 1.0)
        learned_r = np.mean(J)/self.env.horizon
        return learned_r, raw_r, learned_r - raw_r



import numpy as np
import pandas as pd 
import pickle 
import torch.nn as nn
from itertools import chain
from sklearn.ensemble import RandomForestClassifier

from mushroom_rl.environments import Environment, MDPInfo
from mushroom_rl.utils import spaces

from RL_for_reco.TorchModel import ModelMaker, FlexibleTorchModel

class Item_Reco(Environment):
    def __init__(self, items, gamma, horizon, trans_model_abs_path, item_dist=None):
        # MDP parameters

        # 1) discrete actions: list of item names or representing integers
        # 2) actions on n-dimensional space: list of a pair of min and max values per action
        self.items = items
        self.action_dim = len(self.items)
        if item_dist is None:
            if len(self.items.shape) == 1:
                if 'none' in self.items:
                    self.item_dist = np.zeros(self.action_dim)
                    self.item_dist[1:] = 1/(self.action_dim-1)
                else:
                    self.item_dist = 1/(self.action_dim)
            else:
                self.item_dist = None
        else:
            self.item_dist = item_dist
        self.gamma = gamma    ## discount factor
        self.horizon = horizon    ## time limit to long
        self.trans_model = ModelMaker(FlexibleTorchModel, model_dir_path=trans_model_abs_path)
        self.trans_model_params = self.trans_model.model.state_dict()
        tmp = list(self.trans_model_params.keys())
        key = list(filter(lambda x: '0.weight' in x, tmp))[0]
        self.state_dim = self.trans_model_params[key].shape[1] - self.action_dim
        if 'none' in self.items:
            self.state_dim += 1

        MM_VAL = 100
        self.min_point = np.ones(self.state_dim) * -MM_VAL
        self.max_point = np.ones(self.state_dim) * MM_VAL
        
        if len(self.items.shape) == 1:
            self._discrete_actions = list(range(self.action_dim))
        else:
            self._discrete_actions = None

        # MDP properties
        observation_space = spaces.Box(low=self.min_point, high=self.max_point)
        if len(self.items.shape) == 1:
            action_space = spaces.Discrete(self.action_dim)
        else:
            action_space = spaces.Box(low=self.items[0][0], high=self.items[0][1])
        mdp_info = MDPInfo(observation_space, action_space, gamma, horizon)

        super().__init__(mdp_info)

    def reset(self, state=None):
        if state is None:
            self._state = np.zeros(self.state_dim)
        else:
            self._state = np.array(state)
        return self._state

    def step(self, action):
        if self._discrete_actions is None:
            next_state, reward = self.trans_model.infer(np.concatenate([self._state, action]))
        else:
            if 'none' in self.items:
                action_onehot = np.zeros(self.action_dim-1)
                if action > 0:
                    action_onehot[action-1] = 1.0
            else:
                action_onehot = np.zeros(self.action_dim)
                action_onehot[action] = 1.0
            next_state, reward = self.trans_model.infer(np.concatenate([self._state, action_onehot]))
        
        return next_state, reward, False, {}


def predict_actions(agent, states, items, none_tree, n_jobs=None, labeled=True):
        actions = list(map(lambda x: agent.draw_action(x), np.array(states)))
        actions = np.array(list(chain(*actions)))
        if labeled:
            str_actions = np.array(items)[actions] 
            if 'none' in str_actions:
                none_idx = np.array(range(len(str_actions)))[str_actions == 'none']

                if type(none_tree) == str:
                    try:
                        none_tree_md = pickle.load(open(none_tree, 'rb'))
                    except:
                        rec_idx = np.array(range(len(str_actions)))[str_actions != 'none']
                        none_tree_md = RandomForestClassifier(n_jobs=n_jobs, n_estimators=50, class_weight='balanced', max_features=0.8, max_depth=5, criterion='entropy').fit(np.array(states)[rec_idx], str_actions[rec_idx])
                        pickle.dump(none_tree_md, open(none_tree, 'wb'), 4)
                else:
                    none_tree_md = none_tree

                print_df = pd.DataFrame([pd.value_counts(str_actions).to_dict()])
                none_mapped = none_tree_md.predict(np.array(states)[none_idx])
                str_actions[none_idx] = none_mapped
                print_df = pd.concat([print_df, pd.DataFrame([pd.value_counts(str_actions).to_dict()])])
                print(print_df)
                return str_actions
            else:
                return str_actions
        else:
            return actions


"""Pytorch implementation of Class-Balanced-Loss
   Reference: "Class-Balanced Loss Based on Effective Number of Samples" 
   Authors: Yin Cui and
               Menglin Jia and
               Tsung Yi Lin and
               Yang Song and
               Serge J. Belongie
   https://arxiv.org/abs/1901.05555, CVPR'19.
"""


import numpy as np
import torch
import torch.nn.functional as F



def focal_loss(labels, logits, alpha, gamma):
    """Compute the focal loss between `logits` and the ground truth `labels`.
    Focal loss = -alpha_t * (1-pt)^gamma * log(pt)
    where pt is the probability of being classified to the true class.
    pt = p (if true class), otherwise pt = 1 - p. p = sigmoid(logit).
    Args:
      labels: A float tensor of size [batch, num_classes].
      logits: A float tensor of size [batch, num_classes].
      alpha: A float tensor of size [batch_size]
        specifying per-example weight for balanced cross entropy.
      gamma: A float scalar modulating loss from hard and easy examples.
    Returns:
      focal_loss: A float32 scalar representing normalized total loss.
    """    
    BCLoss = F.binary_cross_entropy_with_logits(input = logits, target = labels,reduction = "none")

    if gamma == 0.0:
        modulator = 1.0
    else:
        modulator = torch.exp(-gamma * labels * logits - gamma * torch.log(1 + 
            torch.exp(-1.0 * logits)))

    loss = modulator * BCLoss

    weighted_loss = alpha * loss
    focal_loss = torch.sum(weighted_loss)

    focal_loss /= torch.sum(labels)
    return focal_loss



def CB_loss(labels, logits, samples_per_cls, no_of_classes, loss_type, beta, gamma):
    """Compute the Class Balanced Loss between `logits` and the ground truth `labels`.
    Class Balanced Loss: ((1-beta)/(1-beta^n))*Loss(labels, logits)
    where Loss is one of the standard losses used for Neural Networks.
    Args:
      labels: A int tensor of size [batch].
      logits: A float tensor of size [batch, no_of_classes].
      samples_per_cls: A python list of size [no_of_classes].
      no_of_classes: total number of classes. int
      loss_type: string. One of "sigmoid", "focal", "softmax".
      beta: float. Hyperparameter for Class balanced loss.
      gamma: float. Hyperparameter for Focal loss.
    Returns:
      cb_loss: A float tensor representing class balanced loss
    """
    effective_num = 1.0 - np.power(beta, samples_per_cls)
    weights = (1.0 - beta) / np.array(effective_num)
    weights = weights / np.sum(weights) * no_of_classes

    labels_one_hot = F.one_hot(labels, no_of_classes).float()

    weights = torch.tensor(weights).float()
    weights = weights.unsqueeze(0)
    weights = weights.repeat(labels_one_hot.shape[0],1) * labels_one_hot
    weights = weights.sum(1)
    weights = weights.unsqueeze(1)
    weights = weights.repeat(1,no_of_classes)

    if loss_type == "focal":
        cb_loss = focal_loss(labels_one_hot, logits, weights, gamma)
    elif loss_type == "sigmoid":
        cb_loss = F.binary_cross_entropy_with_logits(input = logits,target = labels_one_hot, weights = weights)
    elif loss_type == "softmax":
        pred = logits.softmax(dim = 1)
        cb_loss = F.binary_cross_entropy(input = pred, target = labels_one_hot, weight = weights)
    return cb_loss

'''
if __name__ == '__main__':
    no_of_classes = 5
    logits = torch.rand(10,no_of_classes).float()
    labels = torch.randint(0,no_of_classes, size = (10,))
    beta = 0.9999
    gamma = 2.0
    samples_per_cls = [2,3,1,2,2]
    loss_type = "focal"
    cb_loss = CB_loss(labels, logits, samples_per_cls, no_of_classes,loss_type, beta, gamma)
    print(cb_loss)
'''


import setuptools
from os import path

from RL_for_reco import __version__

here = path.abspath(path.dirname(__file__))

air = [
    "torch>=1.4.0",
    "torchvision>=0.5.0",
    "scikit-learn>=0.22.2.post1",
    "scipy>=1.4.1",
    "matplotlib>=3.2.1",
    "mushroom_rl>=1.4.0",
]


setuptools.setup(
    name='RL_for_reco',
    version=__version__,
    description='A Python toolkit of Deep Reinforcement Learning for Structured Data-Oriented Recommendation.',
    url='https://github.com/gowun/RL_for_reco.git',
    author="Gowun Jeong",
    author_email='gowun.jeong@gmail.com',
    license='MIT',
    packages=setuptools.find_packages(),
    python_requires=">=3.6",
    zip_safe=False,
    long_description=open('README.md').read(),
    install_requires=air,
    classifiers=["Programming Language :: Python :: 3",
                 "License :: OSI Approved :: MIT License",
                 "Operating System :: OS Independent",
                 ],
)



import numpy as np
import pandas as pd 
import pickle 

import torch
import torch.nn as nn
import torch.optim as optim
import torch.nn.functional as F

from mushroom_rl.approximators.parametric.torch_approximator import TorchApproximator
from mushroom_rl.utils.torch import zero_grad

class Network_for_Reco(nn.Module):
    def __init__(self, input_shape, output_shape, hidden_dims, mode='q', **kwargs):
        super().__init__()
        # mode = 'q': q-value network
        # mode = 'actor': actor network
        # mode = 'critic': critic network
        self.mode = mode

        self.output_size = output_shape[0]

        self.fully_connected_net = []
        in_size = input_shape[-1]
        for i, next_size in enumerate(hidden_dims):
            hh = nn.Linear(in_features=in_size, out_features=next_size)
            nn.init.xavier_uniform_(hh.weight, gain=nn.init.calculate_gain('relu'))
            in_size = next_size
            self.__setattr__(f'_h{i}', hh)
            self.fully_connected_net.append(hh)

        self.last_layer = nn.Linear(in_features=in_size, out_features=output_shape[0])
        nn.init.xavier_uniform_(self.last_layer.weight, gain=nn.init.calculate_gain('linear'))

    def forward(self, state, action=None):
        if self.mode == 'critic' and action is not None:
            state_action = torch.cat((state.float(), action.float()), dim=1)
            for hh in self.fully_connected_net:
                state_action = F.relu(hh(state_action))
            q = self.last_layer(state_action)

            return torch.squeeze(q)
        else:
            features = torch.squeeze(state.float(), 1).float()
            for hh in self.fully_connected_net:
                features = F.relu(hh(features))
            q = self.last_layer(features)

            if action is None:
                return q
            else:
                action = action.long()
                q_acted = torch.squeeze(q.gather(1, action))

                return q_acted



class TorchApproximator_cuda(TorchApproximator):
    def __init__(self, input_shape, output_shape, network, optimizer=None,
                 loss=None, batch_size=0, n_fit_targets=1, use_cuda=False,
                 reinitialize=False, dropout=False, quiet=True, cuda_num=None, **params):

        self._batch_size = batch_size
        self._reinitialize = reinitialize
        self._use_cuda = use_cuda
        self._dropout = dropout
        self._quiet = quiet
        self._n_fit_targets = n_fit_targets
        self.cuda_num = None if cuda_num is None else f'cuda: {cuda_num}'

        self.network = network(input_shape, output_shape, **params)

        if self._use_cuda:
            if cuda_num is not None:
                self.network.to(torch.device(self.cuda_num))
                #print(f'{self.cuda_num} is launced')
            else:
                self.network.cuda()
        if self._dropout:
            self.network.eval()

        if optimizer is not None:
            self._optimizer = optimizer['class'](self.network.parameters(),
                                                 **optimizer['params'])
        
        self._loss = loss

    def _compute_batch_loss(self, batch, use_weights, kwargs):
        if use_weights:
            weights = torch.from_numpy(batch[-1]).type(torch.float)
            if self._use_cuda:
                if self.cuda_num is None:
                    weights = weights.cuda()
                else:
                    weights = weights.to(torch.device(self.cuda_num))
            batch = batch[:-1]

        if not self._use_cuda:
            torch_args = [torch.from_numpy(x) for x in batch]
        else:
            if self.cuda_num is None:
                torch_args = [torch.from_numpy(x).cuda() for x in batch]
            else:
                torch_args = [torch.from_numpy(x).to(torch.device(self.cuda_num)) for x in batch]

        x = torch_args[:-self._n_fit_targets]

        y_hat = self.network(*x, **kwargs)

        if isinstance(y_hat, tuple):
            output_type = y_hat[0].dtype
        else:
            output_type = y_hat.dtype

        y = [y_i.clone().detach().requires_grad_(False).type(output_type) for y_i
             in torch_args[-self._n_fit_targets:]]

        if self._use_cuda:
            if self.cuda_num is None:
                y = [y_i.cuda() for y_i in y]
            else:
                y = [y_i.to(torch.device(self.cuda_num)) for y_i in y]

        if not use_weights:
            loss = self._loss(y_hat, *y)
        else:
            loss = self._loss(y_hat, *y, reduction='none')
            loss @= weights
            loss = loss / weights.sum()

        return loss

    def set_weights(self, weights):
        """
        Setter.
        Args:
            w (np.ndarray): the set of weights to set.
        """
        if self.cuda_num is None:
            set_weights(self.network.parameters(), weights, self._use_cuda)
        else:
            set_weights(self.network.parameters(), weights, self._use_cuda, self.cuda_num)

    def predict(self, *args, output_tensor=False, **kwargs):
        """
        Predict.
        Args:
            args (list): input;
            output_tensor (bool, False): whether to return the output as tensor
                or not;
            **kwargs (dict): other parameters used by the predict method
                the regressor.
        Returns:
            The predictions of the model.
        """
        if not self._use_cuda:
            torch_args = [torch.from_numpy(x) if isinstance(x, np.ndarray) else x
                          for x in args]
            val = self.network.forward(*torch_args, **kwargs)

            if output_tensor:
                return val
            elif isinstance(val, tuple):
                val = tuple([x.detach().numpy() for x in val])
            else:
                val = val.detach().numpy()
        else:
            if self.cuda_num is None:
                torch_args = [torch.from_numpy(x).cuda()
                            if isinstance(x, np.ndarray) else x.cuda() for x in args]
            else:
                torch_args = [torch.from_numpy(x).to(torch.device(self.cuda_num))
                            if isinstance(x, np.ndarray) else x.to(torch.device(self.cuda_num)) for x in args]
            val = self.network.forward(*torch_args,
                                       **kwargs)

            if output_tensor:
                return val
            elif isinstance(val, tuple):
                val = tuple([x.detach().cpu().numpy() for x in val])
            else:
                val = val.detach().cpu().numpy()

        return val

    def diff(self, *args, **kwargs):
        """
        Compute the derivative of the output w.r.t. ``state``, and ``action``
        if provided.
        Args:
            state (np.ndarray): the state;
            action (np.ndarray, None): the action.
        Returns:
            The derivative of the output w.r.t. ``state``, and ``action``
            if provided.
        """
        if not self._use_cuda:
            torch_args = [torch.from_numpy(np.atleast_2d(x)) for x in args]
        else:
            if self.cuda_num is None:
                torch_args = [torch.from_numpy(np.atleast_2d(x)).cuda() for x in args]
            else:
                torch_args = [torch.from_numpy(np.atleast_2d(x)).to(torch.device(self.cuda_num)) for x in args]

        y_hat = self.network(*torch_args, **kwargs)
        n_outs = 1 if len(y_hat.shape) == 0 else y_hat.shape[-1]
        y_hat = y_hat.view(-1, n_outs)

        gradients = list()
        for i in range(y_hat.shape[1]):
            zero_grad(self.network.parameters())
            y_hat[:, i].backward(retain_graph=True)

            gradient = list()
            for p in self.network.parameters():
                g = p.grad.data.detach().cpu().numpy()
                gradient.append(g.flatten())

            g = np.concatenate(gradient, 0)

            gradients.append(g)

        g = np.stack(gradients, -1)

        return g

def set_weights(parameters, weights, use_cuda, cuda_num=None):
    """
    Function used to set the value of a set of torch parameters given a
    vector of values.
    Args:
        parameters (list): list of parameters to be considered;
        weights (numpy.ndarray): array of the new values for
            the parameters;
        use_cuda (bool): whether the parameters are cuda tensors or not;
    """
    idx = 0
    for p in parameters:
        shape = p.data.shape

        c = 1
        for s in shape:
            c *= s

        w = np.reshape(weights[idx:idx + c], shape)

        if not use_cuda:
            w_tensor = torch.from_numpy(w).type(p.data.dtype)
        else:
            if cuda_num is None:
                w_tensor = torch.from_numpy(w).type(p.data.dtype).cuda()
            else:
                w_tensor = torch.from_numpy(w).type(p.data.dtype).to(torch.device(cuda_num))

        p.data = w_tensor
        idx += c

    assert idx == weights.size


import os
import pickle
import time
import numpy as np
import pandas as pd
from itertools import chain
from sklearn.metrics import mean_squared_error
from sklearn.model_selection import train_test_split

from RL_for_reco.class_balanced_loss import CB_loss

import warnings
warnings.filterwarnings('ignore')

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from torch.utils.data import DataLoader, Dataset, TensorDataset


class AverageMeter(object):
    def __init__(self):
        self.reset()

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count


class TorchDataCreation:
    def __init__(self, data_list, long_yn_list=None):
        self.data_list = [np.array(d).astype(float) for d in data_list]
        self.long_yn_list = long_yn_list
        self.n_data = len(self.data_list)
        self.data_length = len(self.data_list[0])
        self.test_ratio = None
    
    def split_index(self, test_ratio):
        self.test_ratio = test_ratio
        if self.long_yn_list is not None and True in self.long_yn_list:
            y = self.data_list[self.long_yn_list.index(True)]
            self.train_idx, self.test_idx, _, _ = train_test_split(range(self.data_length), y, test_size=test_ratio)
        else:
            self.test_idx = np.random.choice(self.data_length, round(self.data_length*test_ratio), replace=False)
            self.train_idx = np.array(list(set(range(self.data_length)) - set(self.test_idx)))
    
    def get_dataset(self, idx=None):
        torchdata_list = []
        for i, d in enumerate(self.data_list):
            if idx is None:
                tmp = d
            else:
                tmp = d[idx]
            #torchdata_list.append(torch.Tensor(tmp).float())
            
            if self.long_yn_list is None or self.long_yn_list[i] == False:
                torchdata_list.append(torch.from_numpy(tmp).float())
            elif self.long_yn_list[i]:
                torchdata_list.append(torch.LongTensor(tmp).long())
            
        return TensorDataset(*torchdata_list)
        
    def get_dataloader(self, torchdata, **params):
        return DataLoader(dataset=torchdata, **params)
    
    def make_dataloader(self, **params):
        if self.test_ratio is None:
            tr_idx = None
        else:
            tr_idx, ts_idx = self.train_idx, self.test_idx
        self.train_loader = self.get_dataloader(self.get_dataset(tr_idx), **params)
        if self.test_ratio is not None:
            params['shuffle'] = False
            self.test_loader = self.get_dataloader(self.get_dataset(ts_idx), **params)
        

class FlexibleTorchModel(nn.Module):
    def __init__(self, in_dim, hidden_dims, out_dims, drop_rate=0.0, out_class_yns=None):
        super(FlexibleTorchModel, self).__init__()
        self.drop_rate = drop_rate
        self.out_class_yns = out_class_yns
        self.in_dim = in_dim
        self.out_dims = out_dims
        
        self.fully_connected_net = []
        in_size = in_dim
        for i, next_size in enumerate(hidden_dims):
            fc = nn.Linear(in_features=in_size, out_features=next_size)
            in_size = next_size
            self.__setattr__('fc{}'.format(i), fc)
            self.fully_connected_net.append(fc)
            
        self.last_layers = []
        for i, od in enumerate(out_dims):
            last = nn.Linear(in_features=in_size, out_features=od)
            self.__setattr__('last{}'.format(i), last)
            self.last_layers.append(last)

    def forward(self, X):
        for i, fc in enumerate(self.fully_connected_net):
            X = F.dropout(F.relu(fc(X)), p=self.drop_rate)
        outputs = []
        for i, last in enumerate(self.last_layers):
            if self.out_class_yns is None or self.out_class_yns[i] == False:
                outputs.append(F.relu(last(X)))
            elif self.out_class_yns[i]:
                outputs.append(F.softmax(last(X)))
        
        return outputs
    


class ModelMaker:
    def __init__(self, model, cuda_num=None, model_dir_path=None, **model_params):
        if cuda_num is None:
            self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        elif type(cuda_num) == int:
            self.device = torch.device(f'cuda: {cuda_num}')
        elif cuda_num == 'cpu':
            self.device = torch.device('cpu')
        self.model_name = model
        if model_dir_path is None:
            self.model_params = model_params
            self.model = self.model_name(**self.model_params).to(self.device)
        else:
            self.load_model(model_dir_path)
            
        ## for CB_loss
        self.focal_params = None
        
    def load_model(self, model_dir_path, use_cuda=False):
        self.model_params = pickle.load(open(f'{model_dir_path}/model_params.pkl', 'rb'))
        self.model = self.model_name(**self.model_params)
        if not use_cuda:
            self.device = torch.device('cpu')
        print(self.device)
        self.model.load_state_dict(torch.load(f'{model_dir_path}/state_dict.pkl', map_location=self.device))
        self.model.to(self.device)
    
    @torch.no_grad()
    def infer(self, any_data):
        if type(any_data) != torch.Tensor:
            tensor_data = torch.tensor(torch.tensor(np.array(any_data).astype(float)).float())
    
        self.model.eval()
        tmp = self.model.forward(tensor_data)
        inferred = []
        for ifr in tmp:
            ifr = ifr.numpy()
            if len(ifr.shape) == 2 and ifr.shape[1] == 1:
                ifr = np.array(list(chain(*ifr)))
            inferred.append(ifr)
    
        return inferred
    
    def evaluate_model_byMSE(self, data_set):
        pred = self.infer(data_set[0])
        mse_list = []
        for i, y in enumerate(data_set[1:]):
            if len(y.shape) == 1:
                mse_list.append(mean_squared_error(y, pred[i]))
            else:
                mse_list.append(np.mean(list(map(lambda x: mean_squared_error(x[0], x[1]), zip(np.array(y), pred[i])))))
        return mse_list
    
    def set_optimizer(self, lr_rate, w_decay=1e-8):
        self.optimizer = optim.Adam(self.model.parameters(), lr=lr_rate, weight_decay=w_decay)
        
    def focal_loss(self, labels, logits):
        return CB_loss(labels, logits, **self.focal_params)
        
    def set_criterions(self, *metrics):
        self.criterions = metrics
    
    def single_run(self, tensor_datasets, train=False, loss_p=None, **focal_params):
        outputs = self.model(tensor_datasets[0])
                
        losses = []
        for i, out in enumerate(outputs): 
            if self.model.out_class_yns is not None and self.model.out_class_yns[i]:
                oo = tensor_datasets[i+1].long()
            else:
                oo = tensor_datasets[i+1]
            if self.criterions[i] == 'focal':
                self.focal_params = focal_params
                ls = self.focal_loss(oo.cpu(), out.cpu())
            else:
                ls = self.criterions[i](oo, out)
            losses.append(ls)
            
        if loss_p is None:
            loss = sum(np.array(losses))
        else:
            loss = sum(np.array(losses) * np.array(loss_p))
        
        if train:
            self.optimizer.zero_grad()
            loss.backward()
            self.optimizer.step()

        return loss.item()
    
    def print_process(self, i, total_iter, batch_time, loss, epoch=0, train=True):
        if train:
            print('Epoch {0} [{1}/{2}]\t'
                  'Time {batch_time.val:.3f} ({batch_time.avg:.3f})\t'
                  'Loss {loss.val:.4f} ({loss.avg:.4f})\t'.format(epoch + 1, i + 1, total_iter, 
                                                                  batch_time=batch_time, loss=loss))
        else:
            print('Test [{0}/{1}]\t'
                  'Time {batch_time.val:.3f} ({batch_time.avg:.3f})\t'
                  'Loss {loss.val:.4f} ({loss.avg:.4f})\t'.format(i + 1, total_iter, 
                                                                  batch_time=batch_time, loss=loss))

    def train(self, loader, epoch, loss_p=None, print_term=None, **focal_params):
        batch_time = AverageMeter()
        losses = AverageMeter()
        
        self.model.train()
    
        total_iter = len(loader)
        end = time.time()
        
        for i, data in enumerate(loader):
            data_tensor = [ts.to(self.device) for ts in data]
            
            loss = self.single_run(data_tensor, True, loss_p, **focal_params)
            
            batch_time.update(time.time() - end)
            end = time.time()
            
            batch_size = data_tensor[0].size(0)
            losses.update(loss, batch_size)
            
            if print_term is not None and i % print_term == 0:
                self.print_process(i, total_iter, batch_time, losses, epoch, True)
        print('***\t Loss {loss.avg:.4f}'.format(loss=losses))
        
    @torch.no_grad()
    def validate(self, loader, loss_p=None, print_term=None, **focal_params):
        batch_time = AverageMeter()
        losses = AverageMeter()
        
        self.model.eval()
    
        total_iter = len(loader)
        end = time.time()
        
        for i, data in enumerate(loader):
            data_tensor = [ts.to(self.device) for ts in data]
            
            loss = self.single_run(data_tensor, False, loss_p, **focal_params)
            
            batch_time.update(time.time() - end)
            end = time.time()
            
            batch_size = data_tensor[0].size(0)
            losses.update(loss, batch_size)
            
            if print_term is not None and i % print_term == 0:
                self.print_process(i, total_iter, batch_time, losses, 0, False)
        print('***\t Loss {loss.avg:.4f}'.format(loss=losses))
    
    def train_validate(self, train_loader, test_loader, total_epoch, loss_p=None, print_term=None, **focal_params):
        for ep in range(total_epoch):
            self.train(train_loader, ep, loss_p, print_term, **focal_params)
            self.validate(test_loader, loss_p, print_term, **focal_params)
    
    def save_model(self, dir_path, hdfs_dir_path=None):
        os.system(f'rm -r {dir_path}')
        os.system(f'mkdir {dir_path}')
            
        torch.save(self.model.state_dict(), f'{dir_path}/state_dict.pkl')
        pickle.dump(self.model_params, open(f'{dir_path}/model_params.pkl', 'wb'), 4)
        
        if hdfs_dir_path is not None:
            upload_file_to_hdfs(dir_path, hdfs_dir_path)


def upload_file_to_hdfs(local_path, hdfs_dir_path):
    filename = local_path.split('/')[-1]
    exist = os.popen(f'hdfs dfs -ls {hdfs_dir_path}/{filename}').readlines()
    if len(exist) > 0:
        print(exist)
        if '.' in filename:
            os.system(f'hdfs dfs -rm {hdfs_dir_path}/{filename}')
        else:
            os.system(f'hdfs dfs -rm -r {hdfs_dir_path}/{filename}')
    
    os.system(f'hdfs dfs -put {local_path} {hdfs_dir_path}/')
    
    return os.popen(f'hdfs dfs -ls {hdfs_dir_path}/{filename}').readlines()